--[[
	Title: log4lua
	
	This is a simplified implementation of log4j 
	(<http://en.wikipedia.org/wiki/Log4j>). This lets us easily create, manage,
	and manipulate logging.
]]
module( "ULib3", package.seeall )

if LOG4LUA_LUA then return end -- Return if file already loaded
LOG4LUA_LUA = true

--[[
	Class: Logger
	
	The logger class. Similar to log4j, implements the same functions as the 
	log4j Logger class.
	
	*Do not instantiate this class yourself, use <Logger:getLogger()>.*
]]
Logger = inherits_from( nil )

--[[
	Enumeration: level_hierarchy
	
	The hierarchy of our log levels. The most verbose log levels are at the 
	bottom and the least verbose are at the top. Always refer to the levels by
	name and never by the number. As shown below, names are accessible using
	Logger.FATAL, Logger.ERROR, etc.
]]
Logger.level_hierarchy = {
	"FATAL",
	"ERROR",
	"WARN",
	"INFO",
	"DEBUG",
	"TRACE",
}
-- Now we want to put the reverse table directly in Logger.
for i, level in ipairs( Logger.level_hierarchy ) do
	Logger[ level ] = i
end

Logger.level = Logger.ERROR
--[[
	Function: Logger:instantiate
	
	*DO NOT CALL DIRECTLY*. 'Constructor' for Logger. Just sets up some tables 
	we need.
	
	Revisions:
	
		v3.00 - Initial
]]
function Logger:instantiate()
	self.children = {}
	self.appenders = {}
end

local root_logger = Logger()
--[[
	Function: Logger:getRootLogger
	
	Returns:
	
		The root logger from which all other loggers inherit.
		
	Revisions:

		v3.00 - Initial
]]
function Logger:getRootLogger()
	return root_logger
end

--[[
	Function: Logger:getLogger
	
	Parameters:
	
		logger_name - [String] Name of the logger you wish to create. Use dot
		notation for inheritence. IE, "foo.bar" is a child of "foo".
		
	Returns:
	
		The Logger class.
		
	Revisions:

		v3.00 - Initial
]]
function Logger:getLogger( logger_name )
	local logger_hierarchy = string.Explode( ".", logger_name )
	local current = self:getRootLogger()
	
	for i, next_logger in ipairs( logger_hierarchy ) do
		if not current.children[ next_logger ] then
			local new = Logger()
			current.children[ next_logger ] = new
			new.parent = current
		end
		
		current = current.children[ next_logger ]
	end
	
	return current
end

--TODO: doc and get method
function Logger:setLevel( level )
	--TODO: bounds checking
	self.level = level
end


--[[
	Function: Logger:addAppender
	
	Parameters:
	
		appender - [BaseAppender class] Appender instance to add to this logger.
		
	Revisions:

		v3.00 - Initial
]]
function Logger:addAppender( appender )
	if not isClass( appender ) or not appender:isa( BaseAppender ) then
		error( 2, "invalid appender given to Logger:addAppender" )
	end
	
	if not table.HasValue( self.appenders, appender ) then
		table.insert( self.appenders, appender )
	end
end

--[[
	Function: Logger:removeAppender
	
	Parameters:
	
		appender - [BaseAppender class] Appender instance to remove from this 
		logger.
		
	Revisions:

		v3.00 - Initial
]]
function Logger:removeAppender( appender )
	local location = ULib.findInTable( self.appenders, appender )
	if location then
		table.remove( self.appenders, location )
	end
end

--[[
	Function: Logger:getAppenders
	
	Returns:
		
		The appenders used in this Logger instance.
		
	Revisions:

		v3.00 - Initial
]]
function Logger:getAppenders()
	return self.appenders
end

--[[
	Function: Logger:removeAllAppenders
	
	Clears the list of appenders for this Logger instance.
	
	Returns:
	
		The list of appenders that existed before it was cleared.
		
	Revisions:

		v3.00 - Initial
]]
function Logger:removeAllAppenders()
	local old_table = self:getAppenders()
	self.appenders = {}
	return old_table
end

--[[
	Function: Logger:log
	
	The heart of this class, this is how you pass messages to log. You probably
	want to call the aliases instead of log directly, though. All log messages 
	are	passed up the inheritance hierarchy.
	
	Parameters:
	
		level - The logging level to use for this message.
		message - The message to log
		
	Revisions:

		v3.00 - Initial
]]
function Logger:log( level, message )
	if self.parent then self.parent:log( level, message ) end -- Pass up
	
	if self.level < level then -- Not on our level
		return
	end
	
	for _, appender in ipairs( self.appenders ) do
		appender:rawLine( level, message )
	end
end

-- Set up easy aliases for each log level
for i, level in ipairs( Logger.level_hierarchy ) do
	Logger[ level:lower() ] = function( self, message )
		self:log( i, message )
	end
end


BaseAppender = inherits_from( nil )
function BaseAppender:rawLine( level, message )
end

ConsoleAppender = inherits_from( BaseAppender )
function ConsoleAppender:rawLine( level, message )
	Msg( string.format( "[%s] %s\n", Logger.level_hierarchy[ level ], 
		message ) )
end

FileAppender = inherits_from( BaseAppender )
function FileAppender:instantiate( file_name )
	-- TODO: add error on bad file name
	self.file_name = file_name
end

function FileAppender:rawLine( level, message )
	message = string.format( "[%s] %s\n", Logger.level_hierarchy[ level ], 
		message )
	
	-- TODO: Make sure file name is still valid
	file.Write( self.file_name, file.Read( self.file_name ) .. message )
end

Logger:getRootLogger():addAppender( ConsoleAppender() )