--[[
	Title: luaunit
	
	This is a simplified implementation of the Google C++ unit testing framework
	(<http://code.google.com/p/googletest>). See wikipedia for more information
	on unit testing.
	
	Here's a quick example of how to use this unit testing:
	> function square( val )
	>    return val * val
	> end
	>
	> local function testSquare()
	>    -- Test to make sure it works at basic ranges
	>    for i=-10, 10 do
	>       assertEq( square( i ), math.pow( i, 2 ) )
	>    end
	> 
	>    -- Test a large number
	>    local num = 10e10
	>    assertEq( square( num ), math.pow( num, 2 ) )
	> end
	> test( testSquare, "square", "basic values" )
	
	Any function starting with assert is a fatal function that will terminate 
	the test suite, any function starting with expect simply prints the failure 
	but continues anyways.
]]
module( "ULib3", package.seeall )

local gTest_name, gSub_name

-- Returns filename and line number or nil if it can't find it
local function getFileInfo()
	local info_table = debug.getinfo( 6 ) -- 6 levels up
	if info_table.what == "C" then
		return -- We got nothing
	end
	return string.format( "%s:%i", info_table.short_src, info_table.currentline )
end

-- Returns the name of the calling function
local function getCallerName()
	return debug.getinfo( 3 ).name or debug.getinfo( 2 ).name or ""
end

-- Throw an error on fatal tests
local function fatal( msg )
	-- TODO, use logging
	local file_info = getFileInfo()
	if file_info then
		error( string.format( "[FATAL] (%s-%s) %s - %s", gTest_name, gSub_name, msg, file_info ), 0 )
	else
		error( string.format( "[FATAL] (%s-%s) %s", gTest_name, gSub_name, msg ), 0 )
	end
end

-- Print a warning on nonfatal tests
local function nonfatal( msg )
	-- TODO, use logging
	local file_info = getFileInfo()
	if file_info then
		print( string.format( "[FAIL] (%s-%s) %s - %s", gTest_name, gSub_name, msg, file_info ) )
	else
		print( string.format( "[FAIL] (%s-%s) %s", gTest_name, gSub_name, msg ) )
	end
end

-- Just a helper function to redirect traffic
local function fail( msg, is_fatal )
	if is_fatal then
		fatal( msg )
	else
		nonfatal( msg )
	end
end

--[[
	Function: test
	
	Initiates a test suite at a given function.
	
	Parameters:
		
		test_fn - The function that contains the test suite.
		test_name - The name of what you're testing.
		sub_name - More info on what you're testing (IE, 'valid ranges', 
			'error throwing', etc).
]]
function test( test_fn, test_name, sub_name )
	-- TODO, add param checking
	gTest_name = test_name or "<unknown>"
	gSub_name = sub_name or "<unknown>"
	
	local was_success, error_msg = pcall( test_fn )
	if not was_success then
		print( error_msg )
	end
	
	gTest_name = "<unknown>"
	gSub_name = "<unknown>"
end



local function testTrue( condition, is_fatal )
	if not condition then
		fail( "failed " .. getCallerName(), is_fatal )
	end
end

--[[
	Function: assertTrue
	
	Parameters:
	
		condition - The value to assure is true.
]]
function assertTrue( condition )
	testTrue( condition, true )
end

--[[
	Function: expectTrue
	
	Parameters:
	
		condition - The value to assure is true.
]]
function expectTrue( condition )
	testTrue( condition, false )
end



local function testFalse( condition, is_fatal )
	if condition then
		fail( "failed " .. getCallerName(), is_fatal )
	end
end

--[[
	Function: assertFalse
	
	Parameters:
	
		condition - The value to assure is false.
]]
function assertFalse( condition )
	testFalse( condition, true )
end

--[[
	Function: expectFalse
	
	Parameters:
	
		condition - The value to assure is false.
]]
function expectFalse( condition )
	testFalse( condition, false )
end



local function testEq( expected, actual, is_fatal )
	if expected != actual then
		fail( string.format( "failed %s, %s *IS NOT* equal to %s", getCallerName(), tostring( expected ), tostring( actual ) ), is_fatal )
	end
end

--[[
	Function: assertEq
	
	Parameters:
	
		expected - The value you are expecting.
		actual - The value you actually got.
]]
function assertEq( expected, actual )
	testEq( expected, actual, true )
end

--[[
	Function: expectEq
	
	Parameters:
	
		expected - The value you are expecting.
		actual - The value you actually got.
]]
function expectEq( expected, actual )
	testEq( expected, actual, false )
end



local function testNe( val1, val2, is_fatal )
	if val1 == val2 then
		fail( string.format( "failed %s, %s *IS* equal to %s", getCallerName(), tostring( val1 ), tostring( val2 ) ), is_fatal )
	end
end

--[[
	Function: assertNe
	
	Parameters:
	
		val1 - The first value to assure is not equal.
		val2 - The second value that should not equal val1.
]]
function assertNe( val1, val2 )
	testNe( val1, val2, true )
end

--[[
	Function: expectNe
	
	Parameters:
	
		val1 - The first value to assure is not equal.
		val2 - The second value that should not equal val1.
]]
function expectNe( val1, val2 )
	testNe( val1, val2, false )
end



local function testLt( val1, val2, is_fatal )
	if val1 >= val2 then
		fail( string.format( "failed %s, %s *IS NOT* less than %s", getCallerName(), tostring( val1 ), tostring( val2 ) ), is_fatal )
	end	
end

--[[
	Function: assertLt
	
	Parameters:
	
		val1 - The first value to assure is less than val2.
		val2 - The second value.
]]
function assertLt( val1, val2 )
	testLt( val1, val2, true )
end

--[[
	Function: expectLt
	
	Parameters:
	
		val1 - The first value to assure is less than val2.
		val2 - The second value.
]]
function expectLt( val1, val2 )
	testLt( val1, val2, false )
end



local function testLe( val1, val2, is_fatal )
	if val1 > val2 then
		fail( string.format( "failed %s, %s *IS NOT* less than or equal to %s", getCallerName(), tostring( val1 ), tostring( val2 ) ), is_fatal )
	end
end

--[[
	Function: assertLe
	
	Parameters:
	
		val1 - The first value to assure is less than or equal to val2.
		val2 - The second value.
]]
function assertLe( val1, val2 )
	testLe( val1, val2, true )
end

--[[
	Function: expectLe
	
	Parameters:
	
		val1 - The first value to assure is less than or equal to val2.
		val2 - The second value.
]]
function expectLe( val1, val2 )
	testLe( val1, val2, false )
end



local function testGt( val1, val2, is_fatal )
	if val1 <= val2 then
		fail( string.format( "failed %s, %s *IS NOT* greater than %s", getCallerName(), tostring( val1 ), tostring( val2 ) ), is_fatal )
	end
end

--[[
	Function: assertGt
	
	Parameters:
	
		val1 - The first value to assure is greater than val2.
		val2 - The second value.
]]
function assertGt( val1, val2 )
	testGt( val1, val2, true )
end

--[[
	Function: expectGt
	
	Parameters:
	
		val1 - The first value to assure is greater than val2.
		val2 - The second value.
]]
function expectGt( val1, val2 )
	testGt( val1, val2, false )
end




local function testGe( val1, val2, is_fatal )
	if val1 < val2 then
		fail( string.format( "failed %s, %s *IS NOT* greater than or equal to %s", getCallerName(), tostring( val1 ), tostring( val2 ) ), is_fatal )
	end
end

--[[
	Function: assertGe
	
	Parameters:
	
		val1 - The first value to assure is greater than or equal to val2.
		val2 - The second value.
]]
function assertGe( val1, val2 )
	testGe( val1, val2, true )
end

--[[
	Function: expectGe
	
	Parameters:
	
		val1 - The first value to assure is greater than or equal to val2.
		val2 - The second value.
]]
function expectGe( val1, val2 )
	testGe( val1, val2, false )
end



local function testStrCaseEq( expected_str, actual_str, is_fatal )
	if expected_str:lower() != actual_str:lower() then
		fail( string.format( "failed %s, '%s' *IS NOT* equal to '%s' when ignoring case", getCallerName(), expected_str, actual_str ), is_fatal )
	end
end

--[[
	Function: assertStrCaseEq
	
	Parameters:
	
		expected_str - The case-insensitive string you are expecting.
		actual_str - The string you got.
]]
function assertStrCaseEq( expected_str, actual_str )
	testStrCaseEq( expected_str, actual_str, true )
end

--[[
	Function: expectStrCaseEq
	
	Parameters:
	
		expected_str - The case-insensitive string you are expecting.
		actual_str - The string you got.
]]
function expectStrCaseEq( expected_str, actual_str )
	testStrCaseEq( expected_str, actual_str, false )
end



local function testStrCaseNe( str1, str2, is_fatal )
	if str1:lower() == str2:lower() then
		fail( string.format( "failed %s, '%s' *IS* equal to '%s' when ignoring case", getCallerName(), str1, str2 ), is_fatal )
	end
end

--[[
	Function: assertStrCaseNe
	
	Parameters:
	
		str1 - The case-insensitive string you want to assure is not equal to str2.
		str2 - The second string.
]]
function assertStrCaseNe( str1, str2 )
	testStrCaseNe( str1, str2, true )
end

--[[
	Function: expecttStrCaseNe
	
	Parameters:
	
		str1 - The case-insensitive string you want to assure is not equal to str2.
		str2 - The second string.
]]
function expectStrCaseNe( str1, str2 )
	testStrCaseNe( str1, str2, false )
end



local function testTablesEq( expected_table, actual_table, is_fatal )
	if not tablesEqual( expected_table, actual_table ) then
		fail( string.format( "failed %s, tables *ARE NOT* equal", getCallerName() ), is_fatal )
	end
end

--[[
	Function: assertTablesEq
	
	Performs a fully recursive comparison of the tables
	
	Parameters:
	
		expected_table - The table you are expecting
		actual_table - The actual table
]]
function assertTablesEq( expected_table, actual_table )
	testTablesEq( expected_table, actual_table, true )
end

--[[
	Function: expectTablesEq
	
	Performs a fully recursive comparison of the tables
	
	Parameters:
	
		expected_table - The table you are expecting
		actual_table - The actual table
]]
function expectTablesEq( expected_table, actual_table )
	testTablesEq( expected_table, actual_table, false )
end



local function testTablesNe( table1, table2, is_fatal )
	if tablesEqual( table1, table2 ) then
		fail( string.format( "failed %s, tables *ARE* equal", getCallerName() ), is_fatal )
	end
end

--[[
	Function: assertTablesNe
	
	Performs a fully recursive comparison of the tables
	
	Parameters:
	
		table1 - The first table that you don't want to equal the second.
		table2 - The second table.
]]
function assertTablesNe( table1, table2 )
	testTablesNe( table1, table2, true )
end

--[[
	Function: expectTablesNe
	
	Performs a fully recursive comparison of the tables
	
	Parameters:
	
		table1 - The first table that you don't want to equal the second.
		table2 - The second table.
]]
function expectTablesNe( table1, table2 )
	testTablesNe( table1, table2, false )
end



local function testErrorOnCall( fn, is_fatal, ... )
	local was_successful, error_msg = pcall( fn, ... )
	if was_successful then
		fail( string.format( "failed %s, no error", getCallerName() ), is_fatal )
	end
end

--[[
	Function: assertErrorOnCall
	
	Parameters:
	
		fn - The function to call and assure there is an error thrown from.
		... - The list of parameters to pass to the function.
]]
function assertErrorOnCall( fn, ... )
	testErrorOnCall( fn, true, ... )
end

--[[
	Function: expectErrorOnCall
	
	Parameters:
	
		fn - The function to call and assure there is an error thrown from.
		... - The list of parameters to pass to the function.
]]
function expectErrorOnCall( fn, ... )
	testErrorOnCall( fn, false, ... )
end



local function testNoErrorOnCall( fn, is_fatal, ... )
	local was_successful, error_msg = pcall( fn, ... )
	if not was_successful then
		fail( string.format( "failed %s, error '%s'", getCallerName(), error_msg ), is_fatal )
	end
end

--[[
	Function: assertNoErrorOnCall
	
	Parameters:
	
		fn - The function to call and assure there are no errors thrown from.
		... - The list of parameters to pass to the function.
]]
function assertNoErrorOnCall( fn, ... )
	testNoErrorOnCall( fn, true, ... )
end

--[[
	Function: expectNoErrorOnCall
	
	Parameters:
	
		fn - The function to call and assure there are no errors thrown from.
		... - The list of parameters to pass to the function.
]]
function expectNoErrorOnCall( fn, ... )
	testNoErrorOnCall( fn, false, ... )
end
