--[[
	Title: Portable
	
	As the name suggests, this file is for library functions that have no attachment to garrysmod.
]]

module( "ULib3", package.seeall )


--[[
	Function: makePatternSafe

	Makes a string safe for pattern usage, like in string.gsub(). Basically replaces all keywords with % and keyword.

	Parameters:

		str - The string to make pattern safe

	Returns:

		The pattern safe string
]]
function makePatternSafe( str )
         return str:gsub( "([%(%)%.%%%+%-%*%?%[%]%^%$])", "%%%1" )
end


--[[
	Function: stripQuotes

	Trims leading and tailing quotes from a string

	Parameters:

		s - The string to strip

	Returns:

		The stripped string
]]
function stripQuotes( s )
	return s:gsub( "^%s*[\"]*(.-)[\"]*%s*$", "%1" )
end


--[[
	Function: unescapeBackslash

	Converts '\\' to '\'

	Parameters:

		s - The string to convert

	Returns:

		The converted string
]]
function unescapeBackslash( s )
	return s:gsub( "\\\\", "\\" )
end


--[[
	Function: splitArgs

	This is similar to string.Explode( " ", str ) except that it will also obey quotation marks.

	Parameters:

		str - The string to split from

	Returns:

		A table containing the individual arguments

	Example:

		:ULib.splitArgs( "This is a \"Cool sentence to\" make \"split up\"" )

		returns...

		{ "This", "is", "a", "Cool sentence to", "make", "split up" }
		
	Known Bugs:
	
		* A quote at the start of a token is assumed to start a quote and vice versa with quote at end of token.
		* Will not correctly pick up '"this "is" an odd thing"', will parse it as a single argument.
]]
function splitArgs( str )
	if not str:find( "%S" ) then return {} end -- Empty table if we have an empty string
	
	str = string.Trim( str )
	local symbols = ULib.explode( "%s", str )
	local args = {}
	local in_quote = false
	
	for i, word in ipairs( symbols ) do
		local sanitized_word = ULib.stripQuotes( word )
		
		if not in_quote then
			table.insert( args, sanitized_word )
		else
			args[ #args ] = string.format( "%s %s", args[ #args ], sanitized_word )
		end
		
		if word:sub( 1, 1 ) == '"' then
			in_quote = true
		end
		
		if word:sub( -1, -1 ) == '"' and word:sub( -2, -2 ) ~= '\\' then
			in_quote = false
		end
	end
	
	return args
end

local function tableContainsHelper( tabl, desired_val, recursive, checked_tables )
	for _, val in pairs( tabl ) do
		if val == desired_val then
			return true
		end
	
		if recursive and type( val ) == TABLE and not tableContains( checked_tables, val ) then
			table.insert( checked_tables, val )
			if tableContainsHelper( val, desired_val, recursive, checked_tables ) then
				return true
			end
		end
	end

	return false
end

function tableContains( tabl, desired_val, recursive )
	return tableContainsHelper( tabl, desired_val, recursive, {} )
end

local function tablesEqualHelper( table1, table2, checked_tables )
	for key, value in pairs( table1 ) do
		local value2 = table2[ key ]
		if value2 ~= value then
			if type( value2 ) == TABLE and type( value ) == TABLE then
				if not tableContains( checked_tables, { value2, value1 } ) and not tableContains( checked_table, { value1, value2 } ) then
					table.insert( checked_tables, { value2, value1 } )
					if not tablesEqual( value, value2, checked_tables ) then
						return false
					end
				end
			else
				return false
			end
		end
	end
	
	return true
end

function tablesEqual( table1, table2 )
	return tablesEqualHelper( table1, table2, {} )
end


--[[
	Section: Inheritance
]]

--[[ 
	Function: inheritsFrom
	
	Creates a psudeo-inheritance for lua. It will search for variables that do
	not exist in derived 'classes' in the parent 'classes', among other things 
	explained below.
	
	Parameters:
	
		baseClass - The class to derive from. This value *must* either be nil
			or a class created using <inheritsFrom()>.
			
	Returns:
		
		The table of the derived class.
		
	Revisions:
	
		v3.00 - Initial.
		
	Notes:
	
		* Adapted with improvements from a lua-users inheritance tutorial
		<http://lua-users.org/wiki/InheritanceTutorial>.
		* Create using Class:create( ... ) or Class( ... ) (equivalent).
		* Whatever's passed in the '...', above, is passed to 
		derived_class:instantiate(). This allows for a 'constructor'.
		
	See Also:
	
		* <root_class>
		* <root_class:create>
		* <root_class:class>
		* <root_class:superClass>
		* <root_class:instantiate>
		* <root_class:isa>
		
	Example:
	
:b = inherits_from( nil )
:function b:instantiate( ... )
:	print( "base", unpack( arg ) )
:end
:
:d = inherits_from( b )
:function d:instantiate( ... )
:	print( "derived", unpack( arg ) )
:end
:
:b1 = b( "should be base" )
:d1 = d( "should be derived" )
:print( "d1 is d?", d1:isa( d ), "is b?", d1:isa( b ) )
:print( "b1 is d?", b1:isa( d ), "is b?", b1:isa( b ) )

	Output:
	
:base    should be base
:derived should be derived
:d1 is d?        true    is b?   true
:b1 is d?        false   is b?   true
]]
function inheritsFrom( base_class )
	local new_class = {}
	
	-- The meta-table for INSTANCES (IE, created with Class:create())
	local instance_mt = { __index = new_class, class=new_class, base_class=base_class }
	
	-- The meta-table for the root_class (this will only ever have one table associated with it)
	local class_mt = table.Copy( instance_mt ) -- Only a few differences so copy
	class_mt.__index = base_class or root_class -- Use base or our special meta-base
	class_mt.instance_mt = instance_mt -- Need this for root_class:create()
	
	setmetatable( new_class, class_mt )

	return new_class
end


--[[
	Table: root_class
	
	This is a local table that holds our functions that we want *all* classes 
	to have.
]]
root_class = {}


--[[
	Function: root_class:call
	
	This is a utility function used by the metatable __call to resolve Class( ... ) to Class:create( ... ).
	
	Parameters:
		
		parent_table - The table of the caller.
		... - Extra construction parameters, passed to Class:instantiate.
		
	Returns:
	
		The 'class instance'.
		
	Revisions:

		v3.00 - Initial.
]]
function root_class:call( parent_table, ... )
	return parent_table:class():create( unpack( arg ) )
end


--[[
	Function: root_class:create
	
	This is used to create new 'class instances'.
	
	Parameters:
	
		... - Extra construction parameters, passed to Class:instantiate.
	
	Revisions:

		v3.00 - Initial.
]]
function root_class:create( ... )
	local newinst = {}
	setmetatable( newinst, getmetatable( self ).instance_mt )
	newinst:instantiate( unpack( arg ) ) -- 'Constructor'
	return newinst
end


-- Return the class object of the instance
function root_class:class()
	return getmetatable( self ).class
end


-- Return the super class object of the instance
function root_class:superClass()
	base_class = getmetatable( self ).base_class
	return base_class ~= root_class and base_class or nil -- Nil if root class
end


-- We need to make sure this func exists, but can be overridden
function root_class:instantiate()
	end


-- Return true if the caller is an instance of theClass
function root_class:isa( target_class )
	local cur_class = self:class()

	while cur_class and not b_isa do
		if cur_class == target_class then
		 	return true
		else
			cur_class = cur_class:superClass()
		end
	end

	return false
end

function isClass( obj )
	return type( obj ) == "table" and type( obj.isa ) == "function" and obj:isa( root_class ) -- TODO: Does this work?
end